import json


class TimeseriesAnnotationEvent:
    def __init__(self, *,
                 start_time: float,
                 end_time: float,
                 event_type: str,
                 event_id: str
                 ):
        self.start_time = start_time
        self.end_time = end_time
        self.event_type = event_type
        self.event_id = event_id
    def to_dict(self):
        return {
            's': self.start_time,
            'e': self.end_time,
            't': self.event_type,
            'i': self.event_id
        }

class TimeseriesAnnotationEventType:
    def __init__(self, *,
                 event_type: str,
                 label: str,
                 color_index: int
                 ):
        self.event_type = event_type
        self.label = label
        self.color_index = color_index
    def to_dict(self):
        return {
            'event_type': self.event_type,
            'label': self.label,
            'color_index': self.color_index
        }

class TimeseriesAnnotation:
    def __init__(self, *,
                 events: list[TimeseriesAnnotationEvent],
                 event_types: list[TimeseriesAnnotationEventType]
                 ):
        self.events = events
        self.event_types = event_types
    def to_dict(self):
        return {
            'type': 'TimeseriesAnnotation',
            'event_types': [et.to_dict() for et in self.event_types],
            'events': [e.to_dict() for e in self.events]
        }
    def save(self, path: str):
        with open(path, 'w') as f:
            json.dump(self.to_dict(), f)


from typing import Union
import os
import json
import numpy as np
import numpy.typing as npt
import zarr
from .helpers.compute_templates import compute_templates
from .helpers.compute_correlogram_data import compute_correlogram_data
from .helpers.extract_snippets import extract_snippets_in_channel_neighborhood
from .SpikeTrains import SpikeTrains, SpikeTrain
from .Autocorrelograms import create_autocorrelograms
from .CrossCorrelograms import create_cross_correlograms
from .AverageWaveforms import AverageWaveforms, AverageWaveformItem

def create_spike_sorting_digest(*,
    recording,
    sorting,
    output_path: str,
    num_channels_per_neighborhood: int,
    subsample_segment_duration_sec: float,
    subsample_total_duration_sec: float
):
    if not output_path.endswith('.ns-ssd'):
        raise Exception('Output path must end with .ns-ssd')
    
    print(f'Creating spike sorting digest...')
    print(f'Output path: {output_path}')

    # require that output folder does not exist
    if os.path.exists(output_path):
        raise Exception(f'Output folder already exists: {output_path}')
    
    # create the output folder
    os.makedirs(output_path)
    
    import spikeinterface as si

    recording: si.BaseRecording = recording
    sorting: si.BaseSorting = sorting

    unit_ids = serialize_ids(sorting.get_unit_ids())
    channel_ids = serialize_ids(recording.get_channel_ids())

    print('Creating autocorrelograms.ns-acg...')
    create_autocorrelograms(
        sorting=sorting,
        output_path=f'{output_path}/autocorrelograms.ns-acg',
    )

    print(f'Creating spike_trains.ns-spt...')
    spike_trains: list[SpikeTrain] = []
    for unit_id in unit_ids:
        spike_times = sorting.get_unit_spike_train(unit_id, segment_index=0)
        spike_trains.append(SpikeTrain(
            unit_id=unit_id,
            spike_times_sec=spike_times.astype(np.float32)
        ))
    X = SpikeTrains(
        start_time_sec=0,
        end_time_sec=recording.get_num_frames() / recording.get_sampling_frequency(),
        spike_trains=spike_trains
    )
    X.save(f'{output_path}/spike_trains.ns-spt', block_size_sec=300)

    data_zarr_fname = f'{output_path}/data.zarr'
    data_zarr_root_group = zarr.open(data_zarr_fname, mode="w")

    # Extracting subsampled traces
    print('Extracting subsampled traces...')
    subsampled_traces = get_subsampled_traces(
        recording=recording,
        segment_duration_sec=subsample_segment_duration_sec,
        total_duration_sec=subsample_total_duration_sec
    )

    # Writing subsampled traces
    print('Writing subsampled traces...')
    data_zarr_root_group.create_dataset("subsampled_traces", data=subsampled_traces, chunks=(1000000, len(channel_ids)))

    # Creating subsampled sorting
    print('Creating subsampled sorting...')
    subsampled_sorting = get_subsampled_sorting(
        recording=recording,
        sorting=sorting,
        segment_duration_sec=subsample_segment_duration_sec,
        total_duration_sec=subsample_total_duration_sec
    )

    # Creating subsampled_spike_trains.ns-spt
    print('Creating subsampled_spike_trains.ns-spt...')
    spike_trains: list[SpikeTrain] = []
    for unit_id in unit_ids:
        spike_times = subsampled_sorting.get_unit_spike_train(unit_id, segment_index=0)
        spike_trains.append(SpikeTrain(
            unit_id=unit_id,
            spike_times_sec=spike_times.astype(np.float32)
        ))
    X = SpikeTrains(
        start_time_sec=0,
        end_time_sec=subsample_total_duration_sec,
        spike_trains=spike_trains
    )
    X.save(f'{output_path}/subsampled_spike_trains.ns-spt', block_size_sec=300)

    # Computing full templates
    print('Computing full templates...')
    full_templates = compute_templates(traces=subsampled_traces, sorting=subsampled_sorting)

    # Writing full templates
    print('Writing full templates...')
    data_zarr_root_group.create_dataset("full_templates", data=full_templates, chunks=(1000, 1000, len(channel_ids)))

    # determine peak channels from full templates
    print('Determining peak channels...')
    peak_channel_indices = np.argmin(np.min(full_templates, axis=1), axis=1)
    peak_channel_ids = [channel_ids[i] for i in peak_channel_indices]

    # determine channel neighborhoods from channel locations and peak channels
    print('Determining channel neighborhoods...')
    channel_neighborhoods = []
    for i, unit_id in enumerate(unit_ids):
        peak_channel_id = peak_channel_ids[i]
        peak_channel_index = channel_ids.index(peak_channel_id)
        channel_locations = recording.get_channel_locations()
        peak_channel_location = channel_locations[peak_channel_index]
        channel_distances = np.linalg.norm(channel_locations - peak_channel_location, axis=1)
        # use the closest num_channels_per_neighborhood channels
        channel_neighborhood_indices = np.argsort(channel_distances)[:num_channels_per_neighborhood]
        channel_neighborhood = [channel_ids[i] for i in channel_neighborhood_indices]
        channel_neighborhoods.append({
            'unit_id': unit_id,
            'channel_ids': channel_neighborhood,
            'channel_indices': channel_neighborhood_indices,
            'peak_channel_id': peak_channel_id
        })

    avg_waveforms_list: list[AverageWaveformItem] = []
    
    for unit_id in unit_ids:
        print(f'Processing unit {unit_id}...')
        unit_folder = f'{output_path}/units/{unit_id}'
        os.makedirs(unit_folder)

        channel_neighborhood = [x for x in channel_neighborhoods if x['unit_id'] == unit_id][0]
        subsampled_spike_times = subsampled_sorting.get_unit_spike_train(unit_id, segment_index=0)
        subsampled_snippets_in_neighborhood = extract_snippets_in_channel_neighborhood(traces=subsampled_traces, times=subsampled_spike_times, neighborhood=channel_neighborhood["channel_indices"], T1=30, T2=30)

        channel_neighborhood_indices = channel_neighborhood['channel_indices']
        channel_locations_in_neighborhood = np.array(recording.get_channel_locations())[channel_neighborhood_indices]

        average_waveform_in_neighborhood = np.median(subsampled_snippets_in_neighborhood, axis=0)

        # spike amplitudes
        peak_channel_index_in_neighborhood = np.argmin(np.min(average_waveform_in_neighborhood, axis=0))
        subsampled_spike_amplitudes = np.min(subsampled_snippets_in_neighborhood[:, :, peak_channel_index_in_neighborhood], axis=1)

        # write unit_info.json
        unit_info = {
            'channel_neighborhood_ids': channel_neighborhood['channel_ids'],
            'channel_neighborhood_locations': serialize_channel_locations(channel_locations_in_neighborhood),
            'peak_channel_id': channel_neighborhood['peak_channel_id'],
            'num_subsampled_events': len(subsampled_spike_times)
        }
        print(f'  Num. channels in neighborhood: {len(channel_neighborhood["channel_ids"])}')
        print(f'  Peak channel: {channel_neighborhood["peak_channel_id"]}')
        print(f'  Num. subsampled events: {len(subsampled_spike_times)}')
        with open(f'{unit_folder}/unit_info.json', 'w') as f:
            json.dump(unit_info, f, indent=2)
        # open data.zarr
        unit_data_zarr_fname = f'{unit_folder}/data.zarr'
        unit_data_zarr_root_group = zarr.open(unit_data_zarr_fname, mode="w")
        unit_data_zarr_root_group.create_dataset("subsampled_spike_times", data=subsampled_spike_times.astype(np.float32), chunks=(100000,))
        unit_data_zarr_root_group.create_dataset("average_waveform_in_neighborhood", data=average_waveform_in_neighborhood.astype(np.float32), chunks=(1000, 1000))
        unit_data_zarr_root_group.create_dataset("subsampled_snippets_in_neighborhood", data=subsampled_snippets_in_neighborhood.astype(np.float32), chunks=(1000, 100, 50))
        unit_data_zarr_root_group.create_dataset("subsampled_spike_amplitudes", data=subsampled_spike_amplitudes.astype(np.float32), chunks=(10000,))

        avg_waveforms_list.append(AverageWaveformItem(
            unit_id=unit_id,
            channel_ids=channel_neighborhood['channel_ids'],
            waveform=average_waveform_in_neighborhood
        ))
    
    # Writing average_waveforms.awf
    print('Writing average_waveforms.ns-awf...')
    channel_locations_dict = {}
    for i in range(len(channel_ids)):
        channel_locations_dict[str(channel_ids[i])] = recording.get_channel_locations()[i, :].tolist()
    X_avg_waveforms = AverageWaveforms(
        average_waveforms=avg_waveforms_list,
        channel_locations=channel_locations_dict
    )
    X_avg_waveforms.save(f'{output_path}/average_waveforms.ns-awf')
    
    # Computing unit template correlations
    print('Computing unit template correlations...')
    all_correlations = []
    for i, unit_id1 in enumerate(unit_ids):
        template1 = full_templates[i, :, :]
        for j, unit_id2 in enumerate(unit_ids):
            if j <= i:
                continue
            template2 = full_templates[j, :, :]
            correlation = np.corrcoef(template1.ravel(), template2.ravel())[0, 1]
            all_correlations.append({
                'unit_id1': unit_id1,
                'unit_id2': unit_id2,
                'correlation': correlation
            })
    kk = 20
    # Choose the best kk correlations for each unit
    unit_pair_ids = []
    for unit_id in unit_ids:
        correlations = [x for x in all_correlations if x['unit_id1'] == unit_id]
        correlations = sorted(correlations, key=lambda x: x['correlation'], reverse=True)
        for i in range(min(kk, len(correlations))):
            unit_pair_ids.append([correlations[i]['unit_id1'], correlations[i]['unit_id2']])
    print(f'Using {len(unit_pair_ids)} unit pairs for similarity comparison.')

    # Creating cross_correlograms.ccg
    print('Creating cross_correlograms.ns-ccg...')
    create_cross_correlograms(
        sorting=subsampled_sorting,
        unit_pairs=unit_pair_ids,
        output_path=f'{output_path}/cross_correlograms.ns-ccg'
    )

    # create spike_sorting_digest_info.json
    spike_sorting_digest_info = {
        'channel_ids': channel_ids,
        'sampling_frequency': float(recording.get_sampling_frequency()),
        'num_frames': int(recording.get_num_frames()),
        'unit_ids': unit_ids,
        'unit_pair_ids': unit_pair_ids,
        'channel_locations': serialize_channel_locations(recording.get_channel_locations())
    }
    print(f'Num. channels: {len(channel_ids)}')
    print(f'Num. units: {len(unit_ids)}')
    print(f'Num. unit pairs: {len(unit_pair_ids)}')
    print(f'Num. frames: {spike_sorting_digest_info["num_frames"]}')
    with open(f'{output_path}/spike_sorting_digest_info.json', 'w') as f:
        json.dump(spike_sorting_digest_info, f, indent=2)

    print('Done creating spike sorting digest.')

def get_chunk_sizes_and_spacing(*,
    num_frames: int,
    sampling_frequency: float,
    segment_duration_sec: float,
    total_duration_sec: float
):
    if total_duration_sec * sampling_frequency >= num_frames:
        # if the total duration is longer than the recording, then just use the entire recording
        return [num_frames], 0
    
    # use chunks of segment_duration_sec seconds
    chunk_size = int(sampling_frequency * min(segment_duration_sec, total_duration_sec))
    # the number of chunks depends on the total duration
    num_chunks = int(np.ceil(total_duration_sec * sampling_frequency / chunk_size))
    chunk_sizes = [chunk_size for i in range(num_chunks)]
    chunk_sizes[-1] = int(total_duration_sec * sampling_frequency - (num_chunks - 1) * chunk_size)
    if num_chunks == 1:
        # if only 1 chunk, then just use the initial chunk
        return chunk_sizes, 0
    else:
        # the spacing between the chunks
        spacing = int((num_frames - np.sum(chunk_sizes)) / (num_chunks - 1))
        return chunk_sizes, spacing

def get_subsampled_traces(
    recording, *,
    segment_duration_sec: float,
    total_duration_sec: float
) -> npt.NDArray:
    import spikeinterface as si

    recording: si.BaseRecording = recording

    chunk_sizes, spacing = get_chunk_sizes_and_spacing(
        num_frames=recording.get_num_frames(),
        sampling_frequency=recording.sampling_frequency,
        segment_duration_sec=segment_duration_sec,
        total_duration_sec=total_duration_sec
    )

    num_chunks = len(chunk_sizes)
    if num_chunks == 1:
        # if only 1 chunk, then just use the initial chunk
        traces = recording.get_traces(start_frame=0, end_frame=int(total_duration_sec * recording.sampling_frequency))
    else:
        traces_list: list[np.ndarray] = []
        tt = 0
        for i in range(num_chunks):
            start_frame = tt
            end_frame = int(start_frame + chunk_sizes[i])
            traces_list.append(recording.get_traces(start_frame=start_frame, end_frame=end_frame))
            tt += int(chunk_sizes[i] + spacing)
        traces = np.concatenate(traces_list, axis=0)
    return traces

def get_subsampled_sorting(
    sorting, *,
    recording,
    segment_duration_sec: float,
    total_duration_sec: float,
    margin_num_frames: int = 100
):
    import spikeinterface as si
    
    sorting: si.BaseSorting = sorting
    recording: si.BaseRecording = recording

    chunk_sizes, spacing = get_chunk_sizes_and_spacing(
        num_frames=recording.get_num_frames(),
        sampling_frequency=recording.sampling_frequency,
        segment_duration_sec=segment_duration_sec,
        total_duration_sec=total_duration_sec
    )

    spike_trains_dict = {}
    for unit_id in sorting.get_unit_ids():
        st = sorting.get_unit_spike_train(unit_id, segment_index=0)
        spike_times_list = []
        t_offset = 0
        for i in range(len(chunk_sizes)):
            start_frame = int(i * (chunk_sizes[i] + spacing))
            end_frame = int(start_frame + chunk_sizes[i])
            spike_times_list.append(st[(st >= start_frame + margin_num_frames) & (st < end_frame - margin_num_frames)] - start_frame + t_offset)
            t_offset += chunk_sizes[i]
        spike_times = np.concatenate(spike_times_list)
        spike_trains_dict[unit_id] = spike_times

    return si.NumpySorting.from_dict(
        [spike_trains_dict],
        sampling_frequency=recording.sampling_frequency
    )

def serialize_ids(ids: Union[list, npt.NDArray]) -> list:
    return [id if isinstance(id, str) else int(id) for id in ids]

def serialize_channel_locations(channel_locations: npt.NDArray) -> list:
    ret = []
    for m in range(channel_locations.shape[0]):
        ret.append({
            'x': float(channel_locations[m, 0]),
            'y': float(channel_locations[m, 1])
        })
    return ret


from typing import Union, List
import numpy as np
import numpy.typing as npt


def extract_snippets(
    traces: npt.NDArray[np.float32], *,
    times: npt.NDArray[np.int32],
    T1: int,
    T2: int
) -> npt.NDArray[np.float32]:
    M = traces.shape[1]
    L = len(times)

    snippets = np.zeros((L, T1 + T2, M), dtype=np.float32)
    for j in range(L):
        t1 = times[j] - T1
        t2 = times[j] + T2
        if 0 <= t1 and t2 < traces.shape[0]:
            snippets[j] = traces[t1:t2]
    return snippets

def extract_snippets_in_channel_neighborhood(
    traces: npt.NDArray[np.float32], *,
    times: npt.NDArray[np.int32],
    neighborhood: Union[List[int], None],
    T1: int,
    T2: int
) -> np.ndarray:
    L = len(times)

    if neighborhood is None:
        neighborhood = list(range(traces.shape[1]))

    snippets = np.zeros((L, T1 + T2, len(neighborhood)), dtype=np.float32)
    for j in range(L):
        t1 = times[j] - T1
        t2 = times[j] + T2
        snippets[j] = traces[t1:t2][:, neighborhood]
            
    return snippets


import json
import numpy as np
from typing import List, Union
from ._serialize import _serialize


class TGDataset:
    def __init__(self, *, name: str, data: dict) -> None:
        self._name = name
        self._data = data
    def to_dict(self, *, start_time_sec: float, end_time_sec: float):
        inds = np.where((self._data['t'] >= start_time_sec) & (self._data['t'] <= end_time_sec))[0]
        return {
            'name': self._name,
            'data': {
                't': self._data['t'][inds],
                'y': self._data['y'][inds]
            }
        }

class TGSeries:
    def __init__(self, *, type: str, dataset: str, encoding: dict, attributes: dict, title: str='') -> None:
        self._type = type
        self._dataset = dataset
        self._encoding = encoding
        self._attributes = attributes
        self._title = title
    def to_dict(self):
        return {
            'type': self._type,
            'dataset': self._dataset,
            'encoding': self._encoding,
            'attributes': self._attributes,
            'title': self._title
        }

class TimeseriesGraph:
    def __init__(self, *,
        legend_opts: Union[None, dict]=None,
        y_range: Union[List[float], None]=None,
        hide_x_gridlines: Union[bool, None]=None,
        hide_y_gridlines: Union[bool, None]=None,
        hide_toolbar: bool=False
    ) -> None:
        self.type = 'TimeseriesGraph'
        self._datasets = []
        self._series = []
        self._legend_opts = legend_opts
        self._y_range = y_range
        # time_offset is used to allow float64 type in the time arrays
        self._time_offset = None
        self._hide_x_gridlines = hide_x_gridlines
        self._hide_y_gridlines = hide_y_gridlines
        self._hide_toolbar = hide_toolbar
    def add_line_series(self, *,
            name: str,
            t: np.array,
            y: np.array,
            color: str,
            width: Union[None, int]=None,
            dash: Union[None, List[int]]=None
        ):
        # allow float64 for time array
        t = self._handle_time_offset_t(t)

        if t.dtype == np.float64:
            raise Exception('Cannot handle float64 datatype for t parameter in add_line_series')
        if y.dtype == np.float64:
            raise Exception('Cannot handle float64 datatype for y parameter in add_line_series')

        attributes = {'color': color}
        if width is not None:
            attributes['width'] = width
        if dash is not None:
            attributes['dash'] = dash
        self._add_series(type='line', name=name, t=t, y=y, attributes=attributes)
        return self
    def add_marker_series(self, *,
            name: str,
            t: np.array,
            y: np.array,
            color: str,
            radius: Union[None, int]=None,
            shape: Union[None, str]=None
        ):
        # allow float64 for time array
        t = self._handle_time_offset_t(t)
        
        if t.dtype == np.float64:
            raise Exception('Cannot handle float64 datatype for t parameter in add_marker_series')
        if y.dtype == np.float64:
            raise Exception('Cannot handle float64 datatype for y parameter in add_marker_series')
        attributes = {'color': color}
        if radius is not None:
            attributes['radius'] = radius
        if shape is not None:
            attributes['shape'] = shape
        self._add_series(type='marker', name=name, t=t, y=y, attributes=attributes)
        return self
    def add_dataset(self, ds: TGDataset):
        self._datasets.append(ds)
        return self
    def add_series(self, s: TGSeries):
        self._series.append(s)
        return self
    def save(self, fname: str, block_size_sec: float=300):
        if not fname.endswith('.ns-tgr'):
            raise Exception('TimeseriesGraph::save fname argument must end with .ns-tgr')
        
        start_time_sec = float(np.min([ds._data['t'][0] for ds in self._datasets]))
        end_time_sec = float(np.max([ds._data['t'][-1] for ds in self._datasets]))
        
        block_starts = np.arange(start_time_sec, end_time_sec, block_size_sec)
        block_ends = np.concatenate([block_starts[1:], [end_time_sec]])

        # split into blocks
        json_lines: list[str] = []
        header = {
            'type': 'TimeseriesGraph',
            'startTimeSec': start_time_sec,
            'endTimeSec': end_time_sec,
            'blockSizeSec': block_size_sec,
            'numBlocks': len(block_starts),
            'series': [s.to_dict() for s in self._series]
        }
        if self._time_offset is not None:
            header['timeOffset'] = self._time_offset
        if self._legend_opts is not None:
            header['legendOpts'] = self._legend_opts
        if self._y_range is not None:
            assert len(self._y_range) == 2
            header['yRange'] = self._y_range
        if (self._hide_x_gridlines is not None) or (self._hide_y_gridlines is not None):
            header['gridlineOpts'] = {
                'hideX': True if self._hide_x_gridlines is True else False,
                'hideY': True if self._hide_y_gridlines is True else False
            }
        if self._hide_toolbar:
            header['hideToolbar'] = self._hide_toolbar
        json_lines.append(json.dumps(_serialize(header)))

        for block_start, block_end in zip(block_starts, block_ends):
            x = {
                'startTimeSec': block_start,
                'endTimeSec': block_end,
                'datasets': [
                    ds.to_dict(start_time_sec=block_start, end_time_sec=block_end)
                    for ds in self._datasets
                ]
            }
            json_lines.append(json.dumps(_serialize(x)))
        
        json_line_sizes = [len(line) for line in json_lines]

        with open(fname, 'w') as f:
            f.write(json.dumps(json_line_sizes) + '\n')
            for line in json_lines:
                f.write(line + '\n')

    def _add_series(self, *, type: str, name: str, t: np.ndarray, y: np.ndarray, attributes: dict):
        if t.ndim != 1:
            print('WARNING: TimeseriesGraph::_add_series t argument is not 1D array. Using squeeze.')
            t = np.squeeze(t)
        if y.ndim != 1:
            print('WARNING: TimeseriesGraph::_add_series y argument is not 1D array. Using squeeze.')
            y = np.squeeze(y)
        ds = TGDataset(
            name=name,
            data={
                't': t,
                'y': y
            }
        )
        s = TGSeries(
            type=type,
            encoding={'t': 't', 'y': 'y'},
            dataset=name,
            attributes=attributes,
            title=name
        )
        self.add_dataset(ds)
        self.add_series(s)
    def _handle_time_offset_t(self, t: np.array):
        if t.dtype == np.float64:
            # We have a float64, let's see if we have a time offset
            if self._time_offset is None:
                # we don't, let's make one
                self._time_offset = t[0]
        if self._time_offset is not None:
            # if we have a time offset, let's subtract it
            t = (t - self._time_offset)
            if t.dtype == np.float64:
                # if we have a float64, now that we've subtracted the time offset, it's safe to use float32
                t = t.astype(np.float32)
        return t



import numpy.typing as npt
import numpy as np
import time
from .extract_snippets import extract_snippets


def compute_templates(*, traces: npt.NDArray[np.float32], sorting):
    unit_ids = sorting.unit_ids
    K = len(unit_ids)
    M = traces.shape[1]
    T1 = 30
    T2 = 30
    T = T1 + T2
    print('Compute templates')
    templates = np.zeros((K, T, M), dtype=np.float32)
    timer = time.time()
    for i in range(K):
        elapsed = time.time() - timer
        if elapsed > 5:
            timer = time.time()
            print(f'Computing template for unit {i + 1} of {K}')
        unit_id = unit_ids[i]
        times1 = sorting.get_unit_spike_train(unit_id, segment_index=0)
        snippets1 = extract_snippets(traces, times=times1, T1=T1, T2=T2)
        templates[i] = np.median(snippets1, axis=0)
    return templates


from .TimeseriesGraph import TimeseriesGraph
from .VideoAnnotation import VideoAnnotation, VideoAnnotationNode, VideoAnnotationFrame, VideoAnnotationElement, VideoAnnotationNodeElement, VideoAnnotationEdgeElement
from .SpikeTrains import SpikeTrains, SpikeTrain
from .create_audio_spectrogram import create_audio_spectrogram
from .TimeseriesAnnotation import TimeseriesAnnotation, TimeseriesAnnotationEvent, TimeseriesAnnotationEventType
from .PositionDecodeField import PositionDecodeField, PositionDecodeFieldBin, PositionDecodeFieldFrame
from .Autocorrelograms import Autocorrelograms, AutocorrelogramItem, create_autocorrelograms
from .CrossCorrelograms import CrossCorrelograms, CrossCorrelogramItem
from .AverageWaveforms import AverageWaveforms, AverageWaveformItem
from .create_spike_sorting_digest import create_spike_sorting_digest


import json
import numpy as np
from typing import Union
from ._serialize import _serialize


class SpikeTrain:
    """
    Spike train for a single unit
    """
    def __init__(self,
        unit_id: Union[int, str],
        spike_times_sec: np.array
    ) -> None:
        if spike_times_sec.dtype not in [np.float32, np.int32]:
            raise Exception('spike_times_sec must be float32 or int32')
        self.unit_id = unit_id
        self.spike_times_sec = spike_times_sec

class SpikeTrains:
    def __init__(self, *,
        start_time_sec: float,
        end_time_sec: float,
        spike_trains: list[SpikeTrain]
    ) -> None:
        self._start_time_sec = start_time_sec
        self._end_time_sec = end_time_sec
        self._spike_trains = spike_trains
    def to_single_buffer(self) -> bytes:
        ret = {
            'type': 'SpikeTrains',
            'startTimeSec': self._start_time_sec,
            'endTimeSec': self._end_time_sec,
            'units': [{
                'unitId': a.unit_id,
                'spikeTimesSec': a.spike_times_sec
            } for a in self._spike_trains]
        }
        return json.dumps(_serialize(ret)).encode('utf-8')
    def save(self, fname: str, block_size_sec: float=300):
        if not fname.endswith('.ns-spt'):
            raise Exception('File name must end with .ns-spt')
        
        block_starts = np.arange(self._start_time_sec, self._end_time_sec, block_size_sec)
        block_ends = np.concatenate([block_starts[1:], [self._end_time_sec]])
        
        # split into blocks
        json_lines: list[str] = []
        x = {
            'type': 'SpikeTrains',
            'startTimeSec': self._start_time_sec,
            'endTimeSec': self._end_time_sec,
            'blockSizeSec': block_size_sec,
            'numBlocks': len(block_starts),
            'units': [{
                'unitId': a.unit_id,
                'numSpikes': len(a.spike_times_sec)
            } for a in self._spike_trains]
        }
        json_lines.append(json.dumps(_serialize(x)))

        for block_start, block_end in zip(block_starts, block_ends):
            x = {
                'startTimeSec': block_start,
                'endTimeSec': block_end,
                'units': [
                    {
                        'unitId': a.unit_id,
                        'spikeTimesSec': a.spike_times_sec[
                            (a.spike_times_sec >= block_start) & (a.spike_times_sec < block_end)
                        ]
                    }
                    for a in self._spike_trains
                ]
            }
            json_lines.append(json.dumps(_serialize(x)))
        
        json_lines_sizes = [len(a) for a in json_lines]

        with open(fname, 'w') as f:
            f.write(json.dumps(json_lines_sizes) + '\n')
            for a in json_lines:
                f.write(a + '\n')



import json
from typing import List

import kachery_cloud as kcl
import numpy as np
import simplejson

from ._serialize import _serialize


class PositionDecodeFieldFrame():
    def __init__(self, *, indices: np.ndarray, values: np.ndarray) -> None:
        if type(indices) == list: indices = np.array(indices, dtype=np.uint16)
        if type(values) == list: values = np.array(values, dtype=np.uint16)

        if indices.ndim != 1: raise Exception('indices must be 1D array')
        if indices.dtype != np.uint16: raise Exception('dtype of indices must be np.uint16')
        if values.ndim != 1: raise Exception('values must be 1D array')
        if values.dtype != np.uint16: raise Exception('dtype of values must be np.uint16')
        self.indices = indices
        self.values = values

class PositionDecodeFieldBin():
    def __init__(self, *, x: float, y: float, w: float, h: float) -> None:
        self.x = x
        self.y = y
        self.w = w
        self.h = h
    def to_dict(self):
        return {'x': self.x, 'y': self.y, 'w': self.w, 'h': self.h}

class PositionDecodeField:
    def __init__(self, *,
        bins: list[PositionDecodeFieldBin],
        frames: list[PositionDecodeFieldFrame],
        width: int,
        height: int,
        max_value: float,
        frames_per_second: float
    ) -> None:
        self.bins = bins
        self.frames = frames
        self.max_value = max_value
        self.width = width
        self.height = height
        self.frames_per_second = frames_per_second
    def save(self, fname: str):
        if not fname.endswith('.ns-pdf'):
            raise Exception('File name must end with .ns-pdf')
        
        # split into blocks
        json_lines: list[str] = []
        x = {
            'type': 'PositionDecodeField',
            'width': self.width,
            'height': self.height,
            'frames_per_second': self.frames_per_second,
            'max_value': self.max_value,
            'bins': [bin.to_dict() for bin in self.bins]
        }
        json_lines.append(json.dumps(_serialize(x)))

        for frame in self.frames:
            x = {
                'indices': frame.indices,
                'values': frame.values
            }
            json_lines.append(json.dumps(_serialize(x)))
        
        json_lines_sizes = [len(a) for a in json_lines]

        with open(fname, 'w') as f:
            f.write(json.dumps(json_lines_sizes) + '\n')
            for a in json_lines:
                f.write(a + '\n')


from typing import Tuple, Union
import numpy as np
import neurosift.views as nv
import json
from .views._serialize import _serialize
import os
import json


class RtcsharePlugin:
    def initialize(context):
        context.register_service('pynapple', PynappleService)
        context.register_service('neurosift.sleap', SleapService)
        context.register_service('sleap', SleapService) # temporary alias
        context.register_service('neurosift-nwb-request', NeurosiftNwbRequestService)

class NeurosiftNwbRequestService:
    def handle_query(query: dict, *, dir: str, user_id: Union[str, None]=None) -> Tuple[dict, bytes]:
        from filelock import FileLock
        neurosift_nwb_dir = os.environ['RTCSHARE_DIR'] + '/.neurosift-nwb'
        os.makedirs(neurosift_nwb_dir, exist_ok=True)
        req_fname = f'{neurosift_nwb_dir}/pending-requests.jsonl'
        lock_fname = f'{req_fname}.lock'
        lock = FileLock(lock_fname)
        with lock:
            type0 = query['type']
            if type0 == 'request-meta-nwb':
                nwb_url = query['nwbUrl']
                if not _is_valid_nwb_url(nwb_url):
                    raise Exception(f'Invalid nwbUrl: {nwb_url}')
                with open(req_fname, 'a') as f:
                    f.write(json.dumps({
                        'type': type0,
                        'nwbUrl': nwb_url
                    }) + '\n')
                return {
                    'success': True
                }, b''
            elif type0 == 'request-raster-plot':
                nwb_url = query['nwbUrl']
                path = query['path']
                if not _is_valid_nwb_url(nwb_url):
                    raise Exception(f'Invalid nwbUrl: {nwb_url}')
                if not _is_valid_path(path):
                    raise Exception(f'Invalid path: {path}')
                with open(req_fname, 'a') as f:
                    f.write(json.dumps({
                        'type': type0,
                        'nwbUrl': nwb_url,
                        'path': path
                    }) + '\n')
                return {
                    'success': True
                }, b''
            elif type0 == 'request-autocorrelograms':
                nwb_url = query['nwbUrl']
                path = query['path']
                if not _is_valid_nwb_url(nwb_url):
                    raise Exception(f'Invalid nwbUrl: {nwb_url}')
                if not _is_valid_path(path):
                    raise Exception(f'Invalid path: {path}')
                with open(req_fname, 'a') as f:
                    f.write(json.dumps({
                        'type': type0,
                        'nwbUrl': nwb_url,
                        'path': path
                    }) + '\n')
                return {
                    'success': True
                }, b''
            else:
                raise Exception(f'Unexpected type: {type0}')

def _is_valid_nwb_url(nwb_url: str) -> bool:
    if not nwb_url.startswith('https://') and not nwb_url.startswith('http://'):
        return False
    if len(nwb_url) > 10000:
        return False
    return True


class PynappleService:
    def handle_query(query: dict, *, dir: str, user_id: Union[str, None]=None) -> Tuple[dict, bytes]:
        import pynapple as nap
        from pynapple import TsGroup, TsdFrame
        # todo: authenticate user
        type0 = query['type']
        if type0 == 'session_summary':
            session_uri: str = query['session_uri']
            if session_uri.startswith('$dir'):
                session_uri = f'{dir}/{session_uri[len("$dir"):]}'
            if not session_uri.startswith('rtcshare://'):
                raise Exception(f'Invalid session uri: {session_uri}')
            session_relpath = session_uri[len('rtcshare://'):]
            session_fullpath = os.path.join(os.environ['RTCSHARE_DIR'], session_relpath)
            session = nap.load_session(session_fullpath, 'neurosuite')
            objects = []
            for x in __builtins__['dir'](session):
                v = getattr(session, x)
                if x.startswith('_'):
                    continue
                if isinstance(v, TsGroup):
                    objects.append({
                        'name': x,
                        'type': 'TsGroup'
                    })
                elif isinstance(v, TsdFrame):
                    objects.append({
                        'name': x,
                        'type': 'TsdFrame'
                    })
                elif isinstance(v, dict):
                    objects.append({
                        'name': x,
                        'type': 'dict'
                    })
            return {
                'objects': objects,
                'success': True
            }, b''
        elif type0 == 'get_tsgroup':
            object_name: str = query['object_name']
            session_uri: str = query['session_uri']

            session_fullpath = _full_path_from_uri(session_uri, dir=dir)
            session = nap.load_session(session_fullpath, 'neurosuite')

            obj = getattr(session, object_name)
            spike_trains: list[nv.SpikeTrain] = []
            start_time_sec = None
            end_time_sec = None
            for k in obj.keys():
                spike_times_sec = obj[k].times().astype(np.float32)
                spike_trains.append(nv.SpikeTrain(
                    unit_id=k,
                    spike_times_sec=spike_times_sec
                ))
                if len(spike_times_sec) == 0:
                    continue
                if start_time_sec is None:
                    start_time_sec = spike_times_sec[0]
                else:
                    start_time_sec = min(start_time_sec, spike_times_sec[0])
                if end_time_sec is None:
                    end_time_sec = spike_times_sec[-1]
                else:
                    end_time_sec = max(end_time_sec, spike_times_sec[-1])
            X = nv.SpikeTrains(
                start_time_sec=start_time_sec,
                end_time_sec=end_time_sec,
                spike_trains=spike_trains,
            )
            return {
                'success': True
            }, X.to_single_buffer()
        elif type0 == 'get_tsdframe':
            object_name: str = query['object_name']
            session_fullpath = _full_path_from_uri(session_uri, dir=dir)
            session = nap.load_session(session_fullpath, 'neurosuite')

            obj: TsdFrame = getattr(session, object_name)
            ret = {
                'type': 'TsdFrame',
                'times': obj.times().astype(np.float32),
                'columns': []
            }
            for k in obj.keys():
                ret['columns'].append({
                    'name': k,
                    'values': obj[k].values.astype(np.float32)
                })
            return {
                'success': True
            }, json.dumps(_serialize(ret)).encode('utf-8')
        else:
            raise Exception(f'Unexpected query type: {type0}')

class SleapService:
    def handle_query(query: dict, *, dir: str, user_id: Union[str, None]=None) -> Tuple[dict, bytes]:
        import sleap_io.io.slp as slp
        from sleap_io.model.skeleton import Symmetry
        type0 = query['type']
        if type0 == 'get_skeletons':
            slp_file_uri: str = query['slp_file_uri']
            slp_file_fullpath = _full_path_from_uri(slp_file_uri, dir=dir)
            xx = slp.read_labels(slp_file_fullpath)
            def _skeleton_to_dict(skeleton: slp.Skeleton):
                return {
                    'name': skeleton.name,
                    'node_names': skeleton.node_names,
                    'edge_inds': skeleton.edge_inds,
                    'symmetries': [
                        [
                            node.name
                            for node in sym.nodes 
                        ]
                        for sym in skeleton.symmetries
                    ]
                }
                
            skeletons = [
                _skeleton_to_dict(skeleton)
                for skeleton in xx.skeletons
            ]
            xx.labeled_frames
            return {
                'skeletons': skeletons,
                'success': True
            }, b''

def _full_path_from_uri(uri: str, *, dir: str) -> str:
    if uri.startswith('$dir'):
        uri = f'{dir}/{uri[len("$dir"):]}'
    if not uri.startswith('rtcshare://'):
        raise Exception(f'Invalid uri: {uri}')
    relpath = uri[len('rtcshare://'):]
    fullpath = os.path.join(os.environ['RTCSHARE_DIR'], relpath)
    return fullpath

def _is_valid_path(path: str) -> bool:
    if not path.startswith('/'):
        return False
    if len(path) > 10000:
        return False
    return True


import numpy as np
from typing import Any, Dict, List, Union
import kachery_cloud as kcl
from ._serialize import _serialize


class AverageWaveformItem:
    """
    Single average waveform item (single box)
    """
    def __init__(self,
        unit_id: Union[int, str],
        channel_ids: List[Union[int, str]],
        waveform: np.array,
        waveform_std_dev: Union[None, np.array]=None
    ) -> None:
        self.unit_id = unit_id
        self.channel_ids = channel_ids
        self.waveform = waveform
        self.waveform_std_dev = waveform_std_dev
    def to_dict(self):
        ret = {
            'unitId': self.unit_id,
            'channelIds': self.channel_ids,
            'waveform': self.waveform
        }
        if self.waveform_std_dev is not None:
            ret['waveformStdDev'] = self.waveform_std_dev
        return ret
class AverageWaveforms:
    """
    Average waveforms view
    """
    def __init__(self,
        average_waveforms: List[AverageWaveformItem], *,
        channel_locations: Union[None, Dict[str, Any]] = None
    ) -> None:
        self._average_waveforms = average_waveforms
        self._channel_locations = channel_locations
    def to_dict(self) -> dict:
        ret = {
            'type': 'AverageWaveforms',
            'averageWaveforms': [a.to_dict() for a in self._average_waveforms]
        }
        if self._channel_locations is not None:
            ret['channelLocations'] = self._channel_locations
        return ret
    def save(self, path: str) -> None:
        if not path.endswith('.ns-awf'):
            raise Exception('File name must end with .ns-awf')
        import json
        with open(path, 'w') as f:
            json.dump(_serialize(self.to_dict()), f, indent=2)



import numpy as np
from typing import List, Union
from .helpers.compute_correlogram_data import compute_correlogram_data
from ._serialize import _serialize


class AutocorrelogramItem:
    """
    Single autocorrelogram item (single box)
    """
    def __init__(self,
        unit_id: Union[int, str],
        bin_edges_sec: Union[np.array, List[float]],
        bin_counts: Union[np.array, List[float]]
    ) -> None:
        self.unit_id = unit_id
        self.bin_edges_sec = bin_edges_sec
        self.bin_counts = bin_counts
    def to_dict(self):
        return {
            'unitId': self.unit_id,
            'binEdgesSec': self.bin_edges_sec,
            'binCounts': self.bin_counts
        }

class Autocorrelograms:
    """
    Autocorrelograms view
    """
    def __init__(self,
        autocorrelograms: List[AutocorrelogramItem],
        **kwargs
    ) -> None:
        self._autocorrelograms = autocorrelograms
    def to_dict(self) -> dict:
        ret = {
            'type': 'Autocorrelograms',
            'autocorrelograms': [a.to_dict() for a in self._autocorrelograms]
        }
        return ret
    def save(self, path: str) -> None:
        if not path.endswith('.ns-acg'):
            raise Exception('File name must end with .ns-acg')
        import json
        with open(path, 'w') as f:
            json.dump(_serialize(self.to_dict()), f, indent=2)

def create_autocorrelograms(*, sorting, output_path: str):
    if not output_path.endswith('.ns-acg'):
        raise Exception('File name must end with .ns-acg')
    autocorrelograms: AutocorrelogramItem = []
    unit_ids = sorting.get_unit_ids()
    for unit_id in unit_ids:
        ac = compute_correlogram_data(sorting=sorting, unit_id1=unit_id, window_size_msec=100, bin_size_msec=1)
        item = AutocorrelogramItem(
            unit_id=unit_id,
            bin_edges_sec=ac['bin_edges_sec'],
            bin_counts=ac['bin_counts']
        )
        autocorrelograms.append(item)
    
    X = Autocorrelograms(autocorrelograms)
    X.save(output_path)


import numpy as np
from typing import List, Union, Tuple
from ._serialize import _serialize
from .helpers.compute_correlogram_data import compute_correlogram_data


class CrossCorrelogramItem:
    """
    Single cross correlogram
    """
    def __init__(self,
        unit_id1: Union[int, str],
        unit_id2: Union[int, str],
        bin_edges_sec: Union[np.array, List[float]],
        bin_counts: Union[np.array, List[float]]
    ) -> None:
        self.unit_id1 = unit_id1
        self.unit_id2 = unit_id2
        self.bin_edges_sec = bin_edges_sec
        self.bin_counts = bin_counts
    def to_dict(self):
        return {
            'unitId1': self.unit_id1,
            'unitId2': self.unit_id2,
            'binEdgesSec': self.bin_edges_sec,
            'binCounts': self.bin_counts
        }

class CrossCorrelograms:
    """
    Cross correlograms view
    """
    def __init__(self,
        cross_correlograms: List[CrossCorrelogramItem]
    ) -> None:
        self._cross_correlograms = cross_correlograms
    def to_dict(self) -> dict:
        ret = {
            'type': 'CrossCorrelograms',
            'crossCorrelograms': [a.to_dict() for a in self._cross_correlograms]
        }
        return ret
    def save(self, path: str) -> None:
        if not path.endswith('.ns-ccg'):
            raise Exception('File name must end with .ns-ccg')
        import json
        with open(path, 'w') as f:
            json.dump(_serialize(self.to_dict()), f, indent=2)

def create_cross_correlograms(*, sorting, unit_pairs: list[Tuple[int]], output_path: str):
    if not output_path.endswith('.ns-ccg'):
        raise Exception('File name must end with .ns-ccg')
    cross_correlograms: CrossCorrelogramItem = []
    for unit_pair in unit_pairs:
        ac = compute_correlogram_data(sorting=sorting, unit_id1=unit_pair[0], unit_id2=unit_pair[1], window_size_msec=100, bin_size_msec=1)
        item = CrossCorrelogramItem(
            unit_id1=unit_pair[0],
            unit_id2=unit_pair[1],
            bin_edges_sec=ac['bin_edges_sec'],
            bin_counts=ac['bin_counts']
        )
        cross_correlograms.append(item)
    
    X = CrossCorrelograms(cross_correlograms)
    X.save(output_path)


import os
import shutil
import json
import numpy as np
import zarr


def create_audio_spectrogram(*,
    audio_file_path: str,
    output_path: str
):
    import h5py
    from scipy.io import wavfile
    from matplotlib.pyplot import specgram

    if not output_path.endswith('.ns-asp'):
        raise Exception(f'Output path must end with .ns-asp: {output_path}')
    if os.path.exists(output_path):
        shutil.rmtree(output_path)

    print('Extracting audio signals')
    if audio_file_path.endswith('.h5'):
        audio_sr_hz = _get_audio_sr_from_h5(audio_file_path)
        with h5py.File(audio_file_path, 'r') as f:
            ch1 = np.array(f['ai_channels/ai0'])
            ch2 = np.array(f['ai_channels/ai1'])
            ch3 = np.array(f['ai_channels/ai2'])
            ch4 = np.array(f['ai_channels/ai3'])
            X = np.stack([ch1, ch2, ch3, ch4]).T
    elif audio_file_path.endswith('.wav'):
        audio_sr_hz, X = wavfile.read(audio_file_path)
        if X.ndim == 1:
            X = X.reshape((-1, 1))
        # n_samples = audio.shape[0]
        # n_channels = audio.shape[1]
    else:
        raise Exception(f'Unknown audio file type: {audio_file_path}')
    
    num_channels = X.shape[1]

    print('Computing spectrograms')
    spectrograms = []
    for channel_ind in range(num_channels):
        s, spectrogram_frequencies, spectrogram_times, im = specgram(X[:, channel_ind], NFFT=512, noverlap=256, Fs=audio_sr_hz)
        sr_spectrogram = float(1 / (spectrogram_times[1] - spectrogram_times[0]))
        spectrograms.append(s.T) # Use transpose so we have Nt x Nf
    print(f'Spectrogram sampling rate (Hz): {sr_spectrogram}')
    spectrogram_for_gui = np.log(sum(spectrograms) + 1e-6)

    print('Auto detecting maxval')
    maxval = _auto_detect_spectrogram_maxval(spectrogram_for_gui, sr_spectrogram=sr_spectrogram)
    minval = np.percentile(spectrogram_for_gui, 0.5)
    print(f'Absolute spectrogram max: {np.max(spectrogram_for_gui)}')
    print(f'Auto detected spectrogram max: {maxval}')

    print('Scaling spectogram data')
    
    # Nf x Nt
    scaled_vals = (spectrogram_for_gui - minval) / (maxval - minval + 1e-6) * 255
    scaled_vals = np.clip(scaled_vals, 0, 255)

    spectrogram_for_gui: np.ndarray = scaled_vals.astype(np.uint8)

    # threshold = np.percentile(spectrogram_for_gui[freq_range[0]:freq_range[1]], threshold_pct)
    # print(f'Using threshold: {threshold} ({threshold_pct} pct)')
    # spectrogram_for_gui[spectrogram_for_gui <= threshold] = 0
    
    print(f'Writing {output_path}')
    root_group = zarr.open(output_path, mode="w")
    root_group.create_dataset("spectrogram", data=spectrogram_for_gui, chunks=(50000, len(spectrogram_frequencies)))

    ds_factor = 3
    previous_spectrogram = spectrogram_for_gui
    while ds_factor < len(spectrogram_times):
        spectrogram_for_gui_downsampled = downsample_spectrogram_using_max(previous_spectrogram, ds_factor=3)
        root_group.create_dataset(f"spectrogram_ds{ds_factor}", data=spectrogram_for_gui_downsampled, chunks=(50000, len(spectrogram_frequencies)))
        previous_spectrogram = spectrogram_for_gui_downsampled
        ds_factor *= 3

    root_group.attrs['spectrogram_sr_hz'] = sr_spectrogram
    root_group.attrs['spectrogram_num_samples'] = spectrogram_for_gui.shape[0]
    root_group.attrs['spectrogram_duration_sec'] = spectrogram_for_gui.shape[0] / sr_spectrogram
    root_group.attrs['spectrogram_num_frequencies'] = spectrogram_for_gui.shape[1]
    root_group.create_dataset("frequencies", data=spectrogram_frequencies)
    root_group.create_dataset("times", data=spectrogram_times)

def _auto_detect_spectrogram_maxval(spectrogram: np.array, *, sr_spectrogram: float):
    Nt = spectrogram.shape[0]
    Nf = spectrogram.shape[1]
    chunk_num_samples = int(15 * sr_spectrogram)
    chunk_maxvals: list[float] = []
    i = 0
    while i + chunk_num_samples < Nt:
        chunk = spectrogram[i:i + chunk_num_samples]
        chunk_maxvals.append(np.max(chunk))
        i += chunk_num_samples
    v = np.median(chunk_maxvals)
    return v

def downsample_spectrogram_using_max(spectrogram: np.ndarray, *, ds_factor: int):
    Nt = spectrogram.shape[0]
    Nf = spectrogram.shape[1]
    Nt_ds = Nt // ds_factor
    spectrogram_ds = np.zeros((Nt_ds, Nf), dtype=spectrogram.dtype)
    spectrogram = spectrogram[:Nt_ds * ds_factor, :]
    spectrogram_reshaped = spectrogram.reshape((Nt_ds, ds_factor, Nf))
    spectrogram_ds = np.max(spectrogram_reshaped, axis=1)
    return spectrogram_ds

def _get_audio_sr_from_h5(h5_file: str):
    import h5py
    with h5py.File(h5_file, 'r') as f:
        d = json.loads(f['config'][()].decode('utf-8'))
        audio_sr_hz = d['microphone_sample_rate']
    return audio_sr_hz


import json
from ._serialize import _serialize

class VideoAnnotationElement():
    def __init__(
        self,
        *,
        type: str,
        id: str,
        data: dict
    ) -> None:
        self.type = type
        self.id = id
        self.data = data
    def to_dict(self):
        return {
            't': self.type, # type
            'i': self.id, # id
            **_serialize(self.data) # data
        }

class VideoAnnotationNodeElement(VideoAnnotationElement):
    def __init__(self, id: str, *, x: float, y: float) -> None:
        super().__init__(type='n', id=id, data={'x': x, 'y': y})

class VideoAnnotationEdgeElement(VideoAnnotationElement):
    def __init__(self, id: str, *, id1: str, id2: str) -> None:
        super().__init__(type='e', id=id, data={'i1': id1, 'i2': id2})

class VideoAnnotationFrame():
    def __init__(self, elements: list[VideoAnnotationElement]) -> None:
        self.elements = elements
    def to_dict(self):
        return {
            'e': [el.to_dict() for el in self.elements]
        }

class VideoAnnotationNode():
    def __init__(self, *, id: str, label: str, color_index: int) -> None:
        self.id = id
        self.label = label
        self.color_index = color_index
    def to_dict(self):
        return {
            'id': self.id,
            'label': self.label,
            'colorIndex': self.color_index
        }

class VideoAnnotation:
    def __init__(self, *,
                 width: int,
                 height: int,
                 nodes: list[VideoAnnotationNode],
                 frames: list[VideoAnnotationFrame],
                 frames_per_second: float
                ) -> None:
        self.width = width
        self.height = height
        self.nodes = nodes
        self.frames = frames
        self.frames_per_second = frames_per_second
    def save(self, path: str):
        if not path.endswith('ns-van'):
            raise ValueError('VideoAnnotation.save() requires a path ending in .ns-van')
        first_line = json.dumps({
            'width': self.width,
            'height': self.height,
            'frames_per_second': self.frames_per_second,
            'nodes': [node.to_dict() for node in self.nodes]
        })
        lines: list[str] = [first_line]
        for frame in self.frames:
            lines.append(json.dumps(frame.to_dict()))
        line_lengths = [len(line) for line in lines]
        with open(path, 'w') as f:
            f.write(json.dumps(line_lengths))
            f.write('\n')
            for line in lines:
                f.write(line)
                f.write('\n')


from typing import Union
import numpy as np

def compute_correlogram_data(*, sorting, unit_id1: int, unit_id2: Union[int, None]=None, window_size_msec: float, bin_size_msec: float):
    times1 = sorting.get_unit_spike_train(unit_id=unit_id1, segment_index=0)
    num_bins = int(window_size_msec / bin_size_msec)
    if num_bins % 2 == 0: num_bins = num_bins - 1 # odd number of bins
    num_bins_half = int((num_bins + 1) / 2)
    bin_edges_msec = np.array((np.arange(num_bins + 1) - num_bins / 2) * bin_size_msec, dtype=np.float32)
    bin_counts = np.zeros((num_bins,), dtype=np.int32)
    if unit_id2 is None or unit_id1 == unit_id2:
        # autocorrelogram
        offset = 1
        while True:
            if offset >= len(times1): break
            deltas_msec = (times1[offset:] - times1[:-offset]) / sorting.get_sampling_frequency() * 1000
            deltas_msec = deltas_msec[deltas_msec <= bin_edges_msec[-1]]
            if len(deltas_msec) == 0: break
            for i in range(num_bins_half):
                start_msec = bin_edges_msec[num_bins_half - 1 + i]
                end_msec = bin_edges_msec[num_bins_half + i]
                ct = len(deltas_msec[(start_msec <= deltas_msec) & (deltas_msec < end_msec)])
                bin_counts[num_bins_half - 1 + i] += ct
                bin_counts[num_bins_half - 1 - i] += ct
            offset = offset + 1
    else:
        # cross-correlogram
        times2 = sorting.get_unit_spike_train(segment_index=0, unit_id=unit_id2)
        all_times = np.concatenate((times1, times2))
        all_labels = np.concatenate((1 * np.ones(times1.shape), 2 * np.ones(times2.shape)))
        sort_inds = np.argsort(all_times)
        all_times = all_times[sort_inds]
        all_labels = all_labels[sort_inds]
        offset = 1
        while True:
            if offset >= len(all_times): break
            deltas_msec = (all_times[offset:] - all_times[:-offset]) / sorting.get_sampling_frequency() * 1000

            deltas12_msec = deltas_msec[(all_labels[offset:] == 2) & (all_labels[:-offset] == 1)]
            deltas21_msec = deltas_msec[(all_labels[offset:] == 1) & (all_labels[:-offset] == 2)]
            deltas11_msec = deltas_msec[(all_labels[offset:] == 1) & (all_labels[:-offset] == 1)]
            deltas22_msec = deltas_msec[(all_labels[offset:] == 2) & (all_labels[:-offset] == 2)]

            deltas12_msec = deltas12_msec[deltas12_msec <= bin_edges_msec[-1]]
            deltas21_msec = deltas21_msec[deltas21_msec <= bin_edges_msec[-1]]
            deltas11_msec = deltas11_msec[deltas11_msec <= bin_edges_msec[-1]]
            deltas22_msec = deltas22_msec[deltas22_msec <= bin_edges_msec[-1]]

            if (len(deltas12_msec) + len(deltas21_msec) + len(deltas11_msec) + len(deltas22_msec)) == 0: break
            
            for i in range(num_bins_half):
                start_msec = bin_edges_msec[num_bins_half - 1 + i]
                end_msec = bin_edges_msec[num_bins_half + i]
                ct12 = len(deltas12_msec[(start_msec <= deltas12_msec) & (deltas12_msec < end_msec)])
                ct21 = len(deltas21_msec[(start_msec <= deltas21_msec) & (deltas21_msec < end_msec)])
                bin_counts[num_bins_half - 1 + i] += ct12
                bin_counts[num_bins_half - 1 - i] += ct21
            offset = offset + 1
    return {
        'bin_edges_sec': (bin_edges_msec / 1000).astype(np.float32),
        'bin_counts': bin_counts.astype(np.int32)
    }